package dev.fitko.fitconnect.api.config;

import dev.fitko.fitconnect.api.config.defaults.Environments;
import dev.fitko.fitconnect.api.config.http.HttpConfig;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectInitialisationException;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.ToString;
import lombok.With;
import lombok.experimental.Accessors;

import java.net.URI;
import java.util.HashMap;
import java.util.Map;

import static dev.fitko.fitconnect.api.config.defaults.Environments.getAvailableEnvironmentNames;
import static dev.fitko.fitconnect.api.config.defaults.SchemaConfig.DESTINATION_SCHEMA_VERSION;
import static dev.fitko.fitconnect.api.config.defaults.SchemaConfig.METADATA_SCHEMA_PATH;
import static dev.fitko.fitconnect.api.config.defaults.SchemaConfig.SET_SCHEMA_PATH;

@Getter
@Builder
@Accessors(chain = true)
@ToString
@NoArgsConstructor
@AllArgsConstructor
public class ApplicationConfig {

    private static final String SET_SCHEMA_WRITE_VERSION = "1.2.1";

    private static final String METADATA_SCHEMA_WRITE_VERSION = "1.2.0";

    /***************  Section is customizable via YAML **************/

    private SenderConfig senderConfig;

    private SubscriberConfig subscriberConfig;

    private EnvironmentName activeEnvironment;

    @Builder.Default
    private HttpConfig httpConfig = new HttpConfig();

    @Builder.Default
    private Map<String, String> submissionDataSchemas = new HashMap<>();

    @With
    @Builder.Default
    private Map<EnvironmentName, Environment> environments = Environments.getEnvironmentsAsMap();

    /****************************************************************/

    public boolean isAutoRejectEnabled() {
        return getCurrentEnvironment().getEnableAutoReject();
    }

    public boolean isSkipSubmissionDataValidation() {
        return getCurrentEnvironment().getSkipSubmissionDataValidation();
    }

    public boolean isAllowInsecurePublicKey() {
        return getCurrentEnvironment().getAllowInsecurePublicKey();
    }

    public String getSubmissionBaseUrl() {
        return getCurrentEnvironment().getSubmissionBaseUrls().stream().findFirst().orElseThrow(() -> new FitConnectInitialisationException("No submission base url found, expected at least one"));
    }

    public String getSelfServicePortalBaseUrl() {
        return getCurrentEnvironment().getSelfServicePortalBaseUrl();
    }

    public String getRoutingBaseUrl() {
        return getCurrentEnvironment().getRoutingBaseUrl();
    }

    public String getAuthBaseUrl() {
        return getCurrentEnvironment().getAuthBaseUrl();
    }

    public URI getSetSchemaWriteVersion() {
        return SET_SCHEMA_PATH.getSchemaUriForVersion(SET_SCHEMA_WRITE_VERSION);
    }

    public URI getMetadataSchemaWriteVersion() {
        return METADATA_SCHEMA_PATH.getSchemaUriForVersion(METADATA_SCHEMA_WRITE_VERSION);
    }

    public URI getDestinationSchemaUri() {
        return DESTINATION_SCHEMA_VERSION.getSchemaUri();
    }

    public Environment getCurrentEnvironment() {
        if (activeEnvironment == null){
            throw new FitConnectInitialisationException("Environment must not be null. Please set active environment. Available environments are: " + getAvailableEnvironmentNames());
        }
        if(!environments.containsKey(activeEnvironment)) {
            throw new FitConnectInitialisationException("Environment '" + activeEnvironment.getName() + "' not found. Available environments are: " + getAvailableEnvironmentNames());
        }
        return environments.get(activeEnvironment);
    }
}

