package dev.fitko.fitconnect.client.bootstrap;

import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.FitConnectService;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.SenderConfig;
import dev.fitko.fitconnect.api.config.SubscriberConfig;
import dev.fitko.fitconnect.api.config.resources.BuildInfo;
import dev.fitko.fitconnect.api.config.resources.CertificateConfig;
import dev.fitko.fitconnect.api.config.resources.SchemaResourceConfig;
import dev.fitko.fitconnect.api.domain.schema.SchemaResources;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectInitialisationException;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.api.services.events.CaseService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import dev.fitko.fitconnect.api.services.keys.KeyService;
import dev.fitko.fitconnect.api.services.reply.ReplyService;
import dev.fitko.fitconnect.api.services.routing.RoutingService;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.client.RouterClient;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.sender.ReplyReceiver;
import dev.fitko.fitconnect.client.sender.SubmissionSender;
import dev.fitko.fitconnect.client.subscriber.ReplySender;
import dev.fitko.fitconnect.client.subscriber.SubmissionReceiver;
import dev.fitko.fitconnect.client.util.RessourceLoadingUtils;
import dev.fitko.fitconnect.client.util.ValidDataGuard;
import dev.fitko.fitconnect.core.FitConnectDefaultService;
import dev.fitko.fitconnect.core.auth.DefaultOAuthApiService;
import dev.fitko.fitconnect.core.cases.CaseApiService;
import dev.fitko.fitconnect.core.cases.EventLogVerifier;
import dev.fitko.fitconnect.core.cases.SecurityEventTokenService;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.core.http.HttpClient;
import dev.fitko.fitconnect.core.http.interceptors.ApiRequestInterceptor;
import dev.fitko.fitconnect.core.http.interceptors.UserAgentInterceptor;
import dev.fitko.fitconnect.core.keys.PublicKeyApiService;
import dev.fitko.fitconnect.core.reply.ReplyApiService;
import dev.fitko.fitconnect.core.routing.RouteVerifier;
import dev.fitko.fitconnect.core.routing.RoutingApiService;
import dev.fitko.fitconnect.core.schema.SchemaResourceProvider;
import dev.fitko.fitconnect.core.submission.SubmissionApiService;
import dev.fitko.fitconnect.core.validation.DefaultValidationService;
import okhttp3.Interceptor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.text.ParseException;
import java.util.List;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.client.util.RessourceLoadingUtils.loadYaml;
import static dev.fitko.fitconnect.client.util.RessourceLoadingUtils.readSchemaResourceFromPath;
import static dev.fitko.fitconnect.client.util.RessourceLoadingUtils.readSchemaResourcesFromPaths;

/**
 * Factory that constructs {@link SenderClient}, {@link SubscriberClient} and {@link RouterClient}.
 */
public final class ClientFactory {

    private static final Logger LOGGER = LoggerFactory.getLogger(ClientFactory.class);
    private static final String BUILD_INFO_PATH = "buildinfo.yaml";
    private static final String CERTIFICATE_CONFIG_PATH = "certificate-config.yaml";
    private static final String SCHEMA_CONFIG_PATH = "schema-config.yaml";

    private ClientFactory() {
    }

    /**
     * Create a new {@link SenderClient} to send submissions and receive replies.
     * The client is automatically configured via the provided {@link ApplicationConfig}.
     *
     * @return the sender client
     * @throws FitConnectInitialisationException on errors during the sender client construction
     */
    public static SenderClient createSenderClient(final ApplicationConfig config) throws FitConnectInitialisationException {

        checkNotNull(config);

        LOGGER.info("Initializing sender client ...");
        final SenderConfig senderConfig = config.getSenderConfig();
        final FitConnectService fitConnectService = createFitConnectService(config, senderConfig.getClientId(), senderConfig.getClientSecret());
        final ValidDataGuard dataGuard = new ValidDataGuard(fitConnectService);
        final SubmissionSender submissionSender = new SubmissionSender(config, fitConnectService, dataGuard);
        final ReplyReceiver replyReceiver = new ReplyReceiver(config, fitConnectService);

        return new SenderClient(fitConnectService, submissionSender, replyReceiver);
    }

    /**
     * Create a new {@link SubscriberClient} to receive submissions and send replies.
     * The client is automatically configured via the provided {@link ApplicationConfig}.
     *
     * @return the subscriber client
     * @throws FitConnectInitialisationException on errors during the subscriber client construction
     */
    public static SubscriberClient createSubscriberClient(final ApplicationConfig config) throws FitConnectInitialisationException {

        checkNotNull(config);

        LOGGER.info("Initializing subscriber client ...");
        final SubscriberConfig subscriberConfig = config.getSubscriberConfig();

        final String decryptionKeyPath = getPrivateDecryptionKeyPathFromSubscriber(subscriberConfig);
        LOGGER.info("Reading private decryption key from {}", decryptionKeyPath);
        final String decryptionKey = RessourceLoadingUtils.readKeyFromPath(decryptionKeyPath);
        final RSAKey rsaDecryptionKey = parseRSAKeyFromString(decryptionKey);

        LOGGER.info("Reading private signing key from {} ", config.getSubscriberConfig().getPrivateSigningKeyPath());
        final String signingKey = RessourceLoadingUtils.readKeyFromPath(config.getSubscriberConfig().getPrivateSigningKeyPath());
        final RSAKey rsaSigningKey = parseRSAKeyFromString(signingKey);

        final FitConnectService fitConnectService = createFitConnectService(config, subscriberConfig.getClientId(), subscriberConfig.getClientSecret(), rsaSigningKey);
        final ValidDataGuard dataGuard = new ValidDataGuard(fitConnectService);

        final SubmissionReceiver submissionReceiver = new SubmissionReceiver(fitConnectService, rsaDecryptionKey, config);
        final ReplySender replySender = new ReplySender(config, fitConnectService, dataGuard);

        return new SubscriberClient(fitConnectService, submissionReceiver, replySender);
    }

    /**
     * Create a new {@link RouterClient} to find destinations and services that is automatically configured via a provided {@link ApplicationConfig}.
     *
     * @return the routing client
     * @throws FitConnectInitialisationException on errors during the routing client construction
     */
    public static RouterClient createRouterClient(ApplicationConfig config) throws FitConnectInitialisationException {

        checkNotNull(config);

        LOGGER.info("Initializing router client ...");
        final HttpClient httpClient = getHttpClient(config);
        final SchemaProvider schemaProvider = getSchemaProvider(httpClient, config);
        final ValidationService validator = getValidationService(config, schemaProvider, new HashService());

        final KeyService keyService = new PublicKeyApiService(config, httpClient, validator);
        final RouteVerifier routeVerifier = new RouteVerifier(keyService, validator);
        final RoutingService routingService = new RoutingApiService(httpClient, config.getRoutingBaseUrl());

        return new RouterClient(routingService, routeVerifier);
    }

    private static FitConnectService createFitConnectService(final ApplicationConfig config, final String clientId, final String clientSecret) {
        return createFitConnectService(config, clientId, clientSecret, null);
    }

    private static FitConnectService createFitConnectService(final ApplicationConfig config, final String clientId, final String clientSecret, final RSAKey signingKey) {
        final HttpClient httpClient = getHttpClient(config);
        final SchemaProvider schemaProvider = getSchemaProvider(httpClient, config);
        final MessageDigestService messageDigestService = new HashService();

        final CryptoService cryptoService = new JWECryptoService(messageDigestService);
        final ValidationService validator = getValidationService(config, schemaProvider, messageDigestService);

        final OAuthService authService = new DefaultOAuthApiService(httpClient, clientId, clientSecret, config.getAuthBaseUrl());

        final SubmissionService submissionService = new SubmissionApiService(authService, httpClient, config.getSubmissionBaseUrl());
        final KeyService keyService = new PublicKeyApiService(config, httpClient, authService, submissionService, validator);
        final EventLogVerificationService eventLogVerifier = new EventLogVerifier(keyService, validator);
        final CaseService caseService = new CaseApiService(authService, httpClient, eventLogVerifier, config.getSubmissionBaseUrl());
        final ReplyService replyService = new ReplyApiService(authService, httpClient, config.getSubmissionBaseUrl());

        final SecurityEventService setService = new SecurityEventTokenService(config, validator, signingKey);

        return new FitConnectDefaultService(submissionService, caseService, replyService, cryptoService, validator, keyService, setService);
    }

    private static ValidationService getValidationService(final ApplicationConfig config, final SchemaProvider schemaProvider, final MessageDigestService messageDigestService) {
        final List<String> trustedRootCertificates = RessourceLoadingUtils.loadYaml(CERTIFICATE_CONFIG_PATH, CertificateConfig.class)
                .getTrustedRootCertificates().stream()
                .map(RessourceLoadingUtils::readResourceToString)
                .collect(Collectors.toList());
        LOGGER.info("Initialised trusted root certificates");
        return new DefaultValidationService(config, messageDigestService, schemaProvider, trustedRootCertificates);
    }

    private static HttpClient getHttpClient(final ApplicationConfig config) {
        final BuildInfo buildInfo = loadYaml(BUILD_INFO_PATH, BuildInfo.class);
        final List<Interceptor> interceptors = List.of(new ApiRequestInterceptor(), new UserAgentInterceptor(buildInfo));
        return new HttpClient(config.getHttpConfig(), interceptors);
    }

    private static SchemaProvider getSchemaProvider(final HttpClient httpClient, final ApplicationConfig config) {
        var schemaResourceConfig = loadYaml(SCHEMA_CONFIG_PATH, SchemaResourceConfig.class);

        var metadataSchemas = readSchemaResourcesFromPaths(schemaResourceConfig.getMetadataSchemaPaths());
        var setSchemas = readSchemaResourcesFromPaths(schemaResourceConfig.getSetSchemaPaths());
        var destinationSchema = readSchemaResourceFromPath(schemaResourceConfig.getDestinationSchemaPath());
        var schemaResources = new SchemaResources(setSchemas, metadataSchemas, List.of(destinationSchema), config.getSubmissionDataSchemas());

        return new SchemaResourceProvider(httpClient, schemaResources);
    }

    private static RSAKey parseRSAKeyFromString(final String key) {
        try {
            return RSAKey.parse(JWK.parse(key).toJSONObject());
        } catch (final ParseException | NullPointerException e) {
            LOGGER.error("Could not parse key, please check environment (config.yml)");
            throw new FitConnectInitialisationException("Key could not be parsed", e);
        }
    }

    private static void checkNotNull(final ApplicationConfig config) {
        if (config == null) {
            throw new FitConnectInitialisationException("Application config must not be null");
        }
    }

    private static String getPrivateDecryptionKeyPathFromSubscriber(final SubscriberConfig subscriberConfig) {
        if (subscriberConfig.getPrivateDecryptionKeyPaths().size() != 1) {
            throw new FitConnectInitialisationException("Currently only one configured private key per subscriber is allowed !");
        }
        return subscriberConfig.getPrivateDecryptionKeyPaths().get(0);
    }
}
