package dev.fitko.fitconnect.client.bootstrap;

import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.config.SenderConfig;
import dev.fitko.fitconnect.api.config.SubscriberConfig;
import dev.fitko.fitconnect.api.config.http.HttpConfig;
import dev.fitko.fitconnect.api.config.http.ProxyConfig;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectInitialisationException;
import org.junit.jupiter.api.Test;

import java.util.List;
import java.util.Map;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;

class ClientFactoryTest {

    @Test
    void testMissingConfiguration() {
        assertThrows(FitConnectInitialisationException.class, () -> ClientFactory.createSenderClient(null));
        assertThrows(FitConnectInitialisationException.class, () -> ClientFactory.createSubscriberClient(null));
        assertThrows(FitConnectInitialisationException.class, () -> ClientFactory.createRouterClient(null));
    }

    @Test
    void testSenderClientConstruction() {

        final var envName = new EnvironmentName("DEV");
        final var environments = Map.of(envName, new Environment("https://auth", "", List.of("https://submission.base.net"), "", true, false, false));

        final var sender = new SenderConfig("123", "abc");

        final HttpConfig httpConfig = HttpConfig.builder()
                .proxyConfig(new ProxyConfig("https://proxy.test.de", 8080))
                .build();

        final var senderConfig = ApplicationConfig.builder()
                .httpConfig(httpConfig)
                .senderConfig(sender)
                .environments(environments)
                .activeEnvironment(envName)
                .build();

        assertNotNull(ClientFactory.createSenderClient(senderConfig));
    }

    @Test
    void testSubscriberClientConstruction() {

        final var envName = new EnvironmentName("DEV");
        final var environments = Map.of(envName, new Environment("https://auth", "", List.of("https://submission.base.net"), "", true, false, false));

        final var subscriber = SubscriberConfig.builder()
                .clientSecret("123")
                .clientSecret("abc")
                .privateDecryptionKeyPaths(List.of("src/test/resources/private_decryption_test_key.json"))
                .privateSigningKeyPath("src/test/resources/private_test_signing_key.json")
                .build();

        final HttpConfig httpConfig = HttpConfig.builder()
                .proxyConfig(new ProxyConfig("https://proxy.test.de", 8080))
                .build();

        final var subscriberConfig = ApplicationConfig.builder()
                .httpConfig(httpConfig)
                .environments(environments)
                .subscriberConfig(subscriber)
                .activeEnvironment(envName)
                .build();

        assertNotNull(ClientFactory.createSubscriberClient(subscriberConfig));
    }

    @Test
    void testRouterClientConstruction() {

        final var envName = new EnvironmentName("DEV");
        final var environment = new Environment("", "https://routing.fitko.fitconnect.de", List.of(), "", true, false, false);

        final var senderConfig = new SenderConfig("1234", "abcd");

        final var routingConfig = ApplicationConfig.builder()
                .environments(Map.of(envName, environment))
                .activeEnvironment(envName)
                .senderConfig(senderConfig)
                .build();

        assertNotNull(ClientFactory.createRouterClient(routingConfig));
    }

    @Test
    void testSigningKeyCannotBeParsed() {

        final var envName = new EnvironmentName("DEV");
        final var environments = Map.of(envName, new Environment("https://auth", "", List.of(), "", true, false, false));

        final var subscriberConfig = SubscriberConfig.builder()
                .clientSecret("123")
                .clientSecret("abc")
                .privateDecryptionKeyPaths(List.of("src/test/resources/private_decryption_test_key.json"))
                .privateSigningKeyPath("src/test/resources/invalid_signing_key.json")
                .build();

        final var config = ApplicationConfig.builder()
                .subscriberConfig(subscriberConfig)
                .environments(environments)
                .activeEnvironment(envName)
                .build();

        assertThrows(
                FitConnectInitialisationException.class,
                () -> ClientFactory.createSubscriberClient(config),
                "Expected new ClientFactoryTest() to throw, but nothing was thrown"
        );
    }

    @Test
    void testDecryptionKeyCannotBeParsed() {

        final var envName = new EnvironmentName("DEV");
        final var environments = Map.of(envName, new Environment("https://auth", "", List.of(), "", true, false, false));

        final var subscriberConfigWithoutKey = SubscriberConfig.builder()
                .clientSecret("123")
                .clientSecret("abc")
                .privateDecryptionKeyPaths(List.of("src/test/resources/invalid_private_key.json"))
                .privateSigningKeyPath("src/test/resources/private_test_signing_key.json")
                .build();

        final var config = ApplicationConfig.builder()
                .subscriberConfig(subscriberConfigWithoutKey)
                .environments(environments)
                .activeEnvironment(envName)
                .build();

        assertThrows(
                FitConnectInitialisationException.class,
                () -> ClientFactory.createSubscriberClient(config));
    }

    @Test
    void testMultipleDecryptionKeysNotYetSupported() {

        final var envName = new EnvironmentName("DEV");
        final var environments = Map.of(envName, new Environment("https://auth", "", List.of(), "", true, false, false));

        final var subscriberConfigWithoutKey = SubscriberConfig.builder()
                .clientSecret("123")
                .clientSecret("abc")
                .privateDecryptionKeyPaths(List.of("src/test/resources/private_decryption_test_key.json", "src/test/resources/private_decryption_test_key.json"))
                .privateSigningKeyPath("src/test/resources/private_test_signing_key.json")
                .build();

        final var config = ApplicationConfig.builder()
                .subscriberConfig(subscriberConfigWithoutKey)
                .environments(environments)
                .activeEnvironment(envName)
                .build();

        final FitConnectInitialisationException exception = assertThrows(
                FitConnectInitialisationException.class,
                () -> ClientFactory.createSubscriberClient(config));

        assertThat(exception.getMessage(), containsString("Currently only one configured private key per subscriber is allowed !"));
    }
}