package dev.fitko.fitconnect.core.reply;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.tomakehurst.wiremock.http.ResponseDefinition;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.auth.OAuthToken;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.other.TechnicalError;
import dev.fitko.fitconnect.api.domain.model.reply.AcceptReply;
import dev.fitko.fitconnect.api.domain.model.reply.AnnounceReply;
import dev.fitko.fitconnect.api.domain.model.reply.CreatedReply;
import dev.fitko.fitconnect.api.domain.model.reply.RepliesForPickup;
import dev.fitko.fitconnect.api.domain.model.reply.Reply;
import dev.fitko.fitconnect.api.domain.model.reply.ReplyForPickup;
import dev.fitko.fitconnect.api.domain.model.reply.SentReply;
import dev.fitko.fitconnect.api.domain.model.reply.SubmitReply;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.core.RestEndpointBase;
import dev.fitko.fitconnect.core.http.HttpClient;
import dev.fitko.fitconnect.core.http.MimeTypes;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;

import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static com.github.tomakehurst.wiremock.client.WireMock.containing;
import static com.github.tomakehurst.wiremock.client.WireMock.equalTo;
import static com.github.tomakehurst.wiremock.client.WireMock.get;
import static com.github.tomakehurst.wiremock.client.WireMock.getRequestedFor;
import static com.github.tomakehurst.wiremock.client.WireMock.ok;
import static com.github.tomakehurst.wiremock.client.WireMock.post;
import static com.github.tomakehurst.wiremock.client.WireMock.postRequestedFor;
import static com.github.tomakehurst.wiremock.client.WireMock.put;
import static com.github.tomakehurst.wiremock.client.WireMock.putRequestedFor;
import static com.github.tomakehurst.wiremock.client.WireMock.urlEqualTo;
import static com.github.tomakehurst.wiremock.client.WireMock.verify;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.mockito.Mockito.when;

class ReplyApiServiceTest extends RestEndpointBase {

    private HttpClient httpClient;
    private OAuthService authServiceMock;
    private ReplyApiService underTest;

    @BeforeEach
    public void setUp() {

        final var fakeBaseUrl = "http://localhost:" + wireMockServer.port();
        final ApplicationConfig config = getTestConfig(fakeBaseUrl);

        httpClient = new HttpClient();
        authServiceMock = Mockito.mock(OAuthService.class);
        underTest = new ReplyApiService(authServiceMock, httpClient, config.getSubmissionBaseUrl());
    }

    @Test
    void announceReply() throws JsonProcessingException {

        // Given
        final var caseId = UUID.randomUUID();
        final List<UUID> attachmentIdsToAnnounce = List.of(UUID.randomUUID());

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        final var expectedReply = new CreatedReply();
        expectedReply.setCaseId(caseId);
        expectedReply.setReplyId(UUID.randomUUID());

        final var createReply = new AnnounceReply(caseId, attachmentIdsToAnnounce);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final ResponseDefinition response = wireMockServer.stubFor(
                post(urlEqualTo("/v1/replies"))
                        .withHeader("Authorization", containing("Bearer " + authToken.getAccessToken()))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedReply))
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JSON)
                                .withStatus(201))).getResponse();

        // When
        final CreatedReply createdReply = underTest.announceReply(createReply);

        // Then
        verify(postRequestedFor(urlEqualTo("/v1/replies")).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/json"));
        assertEquals(response.getStatus(), 201);

        assertNotNull(createdReply);
        assertThat(createdReply, is(expectedReply));
    }

    @Test
    void submitReply() throws JsonProcessingException {

        // Given
        final var replyId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var encryptedMetadata = "GfC20IZSSchs27ADalHpRoTZKfXhc7hcMk8Y9V8yTP0jYbmrq6NtEg-QS2O5TQFD9";
        final var encryptedData = "31PBgKjPXHYX1Y6iUcR1sXxSUPjePi8F8P31PBgKjPXHYX1Y6iUcR1sXxSUPjePi8F8P";
        final var attachmentIds = List.of(UUID.randomUUID(), UUID.randomUUID());

        final SentReply expectedSentReply = new SentReply(replyId, caseId, attachmentIds);

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/replies/" + replyId;
        final ResponseDefinition response = wireMockServer.stubFor(
                put(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedSentReply))
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JSON)
                                .withStatus(200))).getResponse();

        final SubmitReply submitReply = new SubmitReply();
        submitReply.setEncryptedMetadata(encryptedMetadata);
        submitReply.setEncryptedData(encryptedData);

        // When
        final SentReply sentReply = underTest.submitReply(replyId, submitReply);

        // Then
        verify(putRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/json"));
        assertEquals(response.getStatus(), 200);

        assertNotNull(sentReply);
        assertThat(sentReply.getReplyId(), is(replyId));

    }

    @Test
    void getReply() throws JsonProcessingException {

        // Given
        final var replyId = UUID.randomUUID();

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        final var expectedReply = new Reply();
        expectedReply.setReplyId(replyId);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/replies/" + replyId;
        final ResponseDefinition response = wireMockServer.stubFor(
                get(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedReply))
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JSON)
                                .withStatus(200))).getResponse();

        // When
        final Reply reply = underTest.getReply(replyId);

        // Then
        verify(getRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/json"));
        assertEquals(response.getStatus(), 200);

        assertNotNull(reply);
        assertThat(reply.getReplyId(), is(replyId));
    }

    @Test
    void getAvailableReplies() throws JsonProcessingException {

        final int offset = 0;
        final int limit = 10;
        final var caseId = UUID.randomUUID();

        final Set<ReplyForPickup> allReplies = IntStream.range(0, 100)
                .mapToObj(i -> new ReplyForPickup(UUID.randomUUID(), caseId))
                .collect(Collectors.toSet());

        final Set<ReplyForPickup> pagedSubmissions = allReplies.stream()
                .skip(offset)
                .limit(limit)
                .collect(Collectors.toSet());

        final RepliesForPickup expectedReplies = new RepliesForPickup();
        expectedReplies.setReplies(pagedSubmissions);
        expectedReplies.setCount(pagedSubmissions.size());
        expectedReplies.setOffset(offset);
        expectedReplies.setTotalCount(allReplies.size());

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/replies?limit=" + limit + "&offset=" + offset;
        final ResponseDefinition response = wireMockServer.stubFor(
                get(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedReplies))
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JSON)
                                .withStatus(200))).getResponse();

        // When
        final RepliesForPickup repliesForPickup = underTest.getAvailableReplies(limit, offset);

        // Then
        verify(getRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/json"));
        assertEquals(response.getStatus(), 200);

        assertNotNull(repliesForPickup);
        assertThat(repliesForPickup.getOffset(), is(offset));
        assertThat(repliesForPickup.getReplies().size(), is(limit));
        assertThat(repliesForPickup.getTotalCount(), is(allReplies.size()));
    }

    @Test
    void acceptReply() {

        // Given
        final var replyId = UUID.randomUUID();

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        final var expectedEvent = "someDummyAcceptEventToken";

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/replies/" + replyId + "/accept";
        final ResponseDefinition response = wireMockServer.stubFor(
                put(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(expectedEvent)
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JOSE)
                                .withStatus(200))).getResponse();

        // When
        final String event = underTest.acceptReply(replyId, new AcceptReply(Collections.emptyList(), new AuthenticationTags()));

        // Then
        verify(putRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/jose"));
        assertEquals(response.getStatus(), 200);

        assertThat(event, is(expectedEvent));
    }

    @Test
    void rejectReply() {

        // Given
        final var replyId = UUID.randomUUID();

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        final var expectedEvent = "someDummyRejectEventToken";

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/replies/" + replyId + "/reject";
        final ResponseDefinition response = wireMockServer.stubFor(
                put(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(expectedEvent)
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JOSE)
                                .withStatus(200))).getResponse();

        // When
        final String event = underTest.rejectReply(replyId, List.of(new TechnicalError()));

        // Then
        verify(putRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/jose"));
        assertEquals(response.getStatus(), 200);

        assertThat(event, is(expectedEvent));
    }

    @Test
    void uploadAttachment() {

        // Given
        final var replyId = UUID.randomUUID();
        final var attachmentId = UUID.randomUUID();
        final var encryptedAttachment = "eyJlbmMiOiJBMjU2R0NNIiwiYWxnIjoiUlNBLU9BRVAtMjU2In0.nlXGAufYH36IABDy0En0";

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/replies/" + replyId + "/attachments/" + attachmentId;
        final ResponseDefinition response = wireMockServer.stubFor(
                put(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(encryptedAttachment)
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JOSE)
                                .withStatus(204))).getResponse();

        // When
        underTest.uploadAttachment(replyId, attachmentId, encryptedAttachment);

        // Then
        verify(putRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/jose"));
        assertEquals(response.getStatus(), 204);
    }

    @Test
    void getAttachment() {

        // Given
        final var replyId = UUID.randomUUID();
        final var attachmentId = UUID.randomUUID();
        final var encryptedAttachment = "eyJlbmMiOiJBMjU2R0NNIiwiYWxnIjoiUlNBLU9BRVAtMjU2In0.nlXGAufYH36IABDy0En0";

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/replies/" + replyId + "/attachments/" + attachmentId;
        final ResponseDefinition response = wireMockServer.stubFor(
                get(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(encryptedAttachment)
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JOSE)
                                .withStatus(200))).getResponse();

        // When
        final String attachment = underTest.getAttachment(replyId, attachmentId);

        // Then
        verify(getRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/jose"));
        assertEquals(response.getStatus(), 200);

        assertNotNull(attachment);
        assertThat(attachment, is(encryptedAttachment));
    }
}