package dev.fitko.fitconnect.core.validation;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWEAlgorithm;
import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.KeyOperation;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.config.defaults.Environments;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.Problem;
import dev.fitko.fitconnect.api.domain.model.event.problems.attachment.AttachmentHashMismatch;
import dev.fitko.fitconnect.api.domain.model.event.problems.attachment.IncorrectAttachmentAuthenticationTag;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataHashMismatch;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataJsonSyntaxViolation;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataXmlSyntaxViolation;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.IncorrectDataAuthenticationTag;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.AttachmentsMismatch;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.IncorrectMetadataAuthenticationTag;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.MetadataJsonSyntaxViolation;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.MetadataSchemaViolation;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.MissingData;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.UnsupportedDataSchema;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.UnsupportedMetadataSchema;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.UnsupportedReplyChannel;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.UnsupportedService;
import dev.fitko.fitconnect.api.domain.model.metadata.AdditionalReferenceInfo;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentForValidation;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.DeMail;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.Email;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.EncryptionPublicKey;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.FitConnect;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.ReplyChannel;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.schema.SchemaResources;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.utils.KeyGenerator;
import dev.fitko.fitconnect.core.http.HttpClient;
import dev.fitko.fitconnect.core.schema.SchemaResourceProvider;
import dev.fitko.fitconnect.jwkvalidator.exceptions.JWKValidationException;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.net.URI;
import java.security.KeyPair;
import java.text.ParseException;
import java.time.Instant;
import java.time.ZonedDateTime;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.contains;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.hasItems;
import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.instanceOf;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

class DefaultValidationServiceTest {

    private DefaultValidationService underTest;
    private MessageDigestService hashService;
    private final HttpClient httpClient = mock(HttpClient.class);
    private SchemaProvider schemaProvider;
    private static Locale defaultLocale;
    private ApplicationConfig config;

    public List<String> trustedRootCertificates;

    @BeforeEach
    void setUp() {

        defaultLocale = Locale.getDefault();
        Locale.setDefault(Locale.ENGLISH);

        final List<String> setSchemas = List.of(getResourceAsString("/set-schemas/set_schema_1.2.1.json"));
        final List<String> metadataSchemas = List.of(getResourceAsString("/metadata-schemas/metadata_schema_1.2.0.json"));
        final List<String> destinationSchemas = List.of(getResourceAsString("/destination-schemas/destination_schema.json"));

        trustedRootCertificates = Stream.of("/trusted-root-certificates/PCA-1-Verwaltung-14.pem",
                        "/trusted-root-certificates/PCA-1-Verwaltung-15.pem",
                        "/trusted-root-certificates/PCA-1-Verwaltung-17.pem",
                        "/trusted-root-certificates/PCA-1-Verwaltung-20.pem",
                        "/trusted-root-certificates/PCA-1-Verwaltung-23.pem")
                .map(this::getResourceAsString)
                .collect(Collectors.toList());

        config = getApplicationConfig(true, false);
        hashService = new HashService();
        schemaProvider = new SchemaResourceProvider(httpClient, new SchemaResources(setSchemas, metadataSchemas, destinationSchemas, config.getSubmissionDataSchemas()));
        underTest = new DefaultValidationService(config, hashService, schemaProvider, trustedRootCertificates);
    }

    @AfterAll
    static void tearDown() {
        Locale.setDefault(defaultLocale);
    }

    @Test
    void testValidKeyWithoutX509CertChain() throws JOSEException {

        //Given
        final RSAKey rsaKey = new RSAKeyGenerator(4096)
                .keyOperations(Set.of(KeyOperation.WRAP_KEY))
                .algorithm(JWEAlgorithm.RSA_OAEP_256)
                .keyID(UUID.randomUUID().toString())
                .generate();

        // When
        final ValidationResult validationResult = underTest.validatePublicKey(rsaKey, KeyOperation.WRAP_KEY);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testInvalidKeyWithWrongAlgorithm() throws JOSEException {

        //Given
        final RSAKey rsaKey = new RSAKeyGenerator(4096)
                .keyOperations(Set.of(KeyOperation.WRAP_KEY))
                .algorithm(JWEAlgorithm.NONE)
                .keyID(UUID.randomUUID().toString())
                .generate();

        // When
        final ValidationResult validationResult = underTest.validatePublicKey(rsaKey, KeyOperation.WRAP_KEY);

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError(), instanceOf(JWKValidationException.class));
        assertThat(validationResult.getError().getMessage(), containsString("JWK with id " + rsaKey.getKeyID() + " is of type wrapKey but has wrong Alg"));
    }

    @Test
    void testInvalidPublicKeyWithInsecureKeyAllowed() throws Exception {

        // Given
        final ApplicationConfig config = getApplicationConfig(true, false);
        final var underTest = new DefaultValidationService(config, hashService, schemaProvider, Collections.emptyList());

        final RSAKey rsaKey = new RSAKeyGenerator(4096)
                .keyOperations(Set.of(KeyOperation.WRAP_KEY))
                .algorithm(JWEAlgorithm.RSA_OAEP_256)
                .keyID(UUID.randomUUID().toString())
                .generate();

        // When
        final ValidationResult validationResult = underTest.validatePublicKey(rsaKey, KeyOperation.WRAP_KEY);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testInvalidPublicKeyWithInsecureKeyNotAllowed() throws Exception {

        // Given
        final ApplicationConfig config = getApplicationConfig(false, false);
        final var underTest = new DefaultValidationService(config, hashService, schemaProvider, trustedRootCertificates);

        final RSAKey rsaKey = new RSAKeyGenerator(4096)
                .keyOperations(Set.of(KeyOperation.WRAP_KEY))
                .algorithm(JWEAlgorithm.RSA_OAEP_256)
                .keyID(UUID.randomUUID().toString())
                .generate();

        // When
        final ValidationResult validationResult = underTest.validatePublicKey(rsaKey, KeyOperation.WRAP_KEY);

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Public key with id '" + rsaKey.getKeyID() + "' does not contain an x5c certificate chain"));
    }


    @Test
    void testValidMetadataSchema() {

        // Given
        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(Collections.emptyList());


        final var metadata = new Metadata();
        metadata.setSchema("https://schema.fitko.de/fit-connect/metadata/1.2.3/metadata.schema.json");
        metadata.setContentStructure(contentStructure);

        // When
        final ValidationResult validationResult = underTest.validateMetadataSchema(metadata);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testInvalidApplicationDateFormat() {

        // Given
        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(Collections.emptyList());

        final var metadata = new Metadata();
        metadata.setSchema("https://schema.fitko.de/fit-connect/metadata/1.2.3/metadata.schema.json");
        metadata.setContentStructure(contentStructure);
        metadata.setAdditionalReferenceInfo(new AdditionalReferenceInfo("ref", "9999-99-999"));

        // When
        final ValidationResult validationResult = underTest.validateMetadataSchema(metadata);

        // Then
        assertFalse(validationResult.isValid());
        assertThat(validationResult.getError().getMessage(), containsString("$.additionalReferenceInfo.applicationDate: does not match the date pattern must be a valid RFC 3339 full-date"));
    }

    @Test
    void testInvalidReplyMetadataSchema() {

        // Given
        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(Collections.emptyList());

        final KeyPair keyPair = KeyGenerator.buildRSAKeyPair(2048);

        final JWK invalidReplyEncryptionKey = KeyGenerator.buildJWK(keyPair, "123", KeyOperation.WRAP_KEY, JWEAlgorithm.RSA_OAEP_256);

        final var metadata = new Metadata();
        metadata.setSchema("https://schema.fitko.de/fit-connect/metadata/1.2.3/metadata.schema.json");
        metadata.setContentStructure(contentStructure);
        metadata.setReplyChannel(ReplyChannel.ofFitConnect(invalidReplyEncryptionKey, List.of("https://process.standard.org")));

        // When
        final ValidationResult validationResult = underTest.validateMetadataSchema(metadata);

        // Then
        assertFalse(validationResult.isValid());
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getError().getMessage(), containsString("$.replyChannel.fitConnect.encryptionPublicKey.kid: must be at least 8 characters long"));
        assertThat(validationResult.getError().getMessage(), containsString("$.replyChannel.fitConnect.encryptionPublicKey.n: must be at least 683 characters long"));

    }

    @Test
    void testValidMetadata() {

        // Given
        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);

        final Hash attachmentOneHash = new Hash();
        attachmentOneHash.setContent(hashService.toHexString(hashService.createHash("someHash".getBytes())));
        attachmentOneHash.setSignatureType(SignatureType.SHA_512);

        final var attachmentOne = new ApiAttachment();
        attachmentOne.setAttachmentId(UUID.randomUUID());
        attachmentOne.setPurpose(Purpose.ATTACHMENT);
        attachmentOne.setMimeType("application/pdf");
        attachmentOne.setHash(attachmentOneHash);

        final Hash attachmentTwoHash = new Hash();
        attachmentTwoHash.setContent(hashService.toHexString(hashService.createHash("someHash".getBytes())));
        attachmentTwoHash.setSignatureType(SignatureType.SHA_512);

        final var attachmentTwo = new ApiAttachment();
        attachmentTwo.setAttachmentId(UUID.randomUUID());
        attachmentTwo.setPurpose(Purpose.ATTACHMENT);
        attachmentTwo.setMimeType("application/pdf");
        attachmentTwo.setHash(attachmentTwoHash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(List.of(attachmentOne, attachmentTwo));

        final KeyPair keyPair = KeyGenerator.buildRSAKeyPair(4096);

        final JWK replyEncryptionKey = KeyGenerator.buildJWK(keyPair, "reply-key:12345678", KeyOperation.WRAP_KEY, JWEAlgorithm.RSA_OAEP_256);

        final ReplyChannel replyChannel = new ReplyChannel();
        replyChannel.setFitConnect(new FitConnect(EncryptionPublicKey.fromJwk(replyEncryptionKey), List.of("https://process.standard.org")));

        final var metadata = new Metadata();
        metadata.setSchema(config.getMetadataSchemaWriteVersion().toString());
        metadata.setContentStructure(contentStructure);
        metadata.setReplyChannel(replyChannel);

        final ServiceType serviceType = new ServiceType();
        serviceType.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");
        submission.setServiceType(serviceType);
        submission.setAttachments(contentStructure.getAttachments().stream().map(ApiAttachment::getAttachmentId).collect(Collectors.toList()));

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(submissionSchema));
        destinationService.setReplyChannels(replyChannel);

        final var destination = new Destination();
        destination.setServices(Set.of(destinationService));

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");
        authenticationTags.setAttachments(Map.of(attachmentOne.getAttachmentId(), "authTag", attachmentTwo.getAttachmentId(), "authTag"));

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testMetadataWithMoreThanOneReplyChannel() {

        // Given
        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);

        final Hash attachmentOneHash = new Hash();
        attachmentOneHash.setContent(hashService.toHexString(hashService.createHash("someHash".getBytes())));
        attachmentOneHash.setSignatureType(SignatureType.SHA_512);

        final var attachmentOne = new ApiAttachment();
        attachmentOne.setAttachmentId(UUID.randomUUID());
        attachmentOne.setPurpose(Purpose.ATTACHMENT);
        attachmentOne.setMimeType("application/pdf");
        attachmentOne.setHash(attachmentOneHash);

        final Hash attachmentTwoHash = new Hash();
        attachmentTwoHash.setContent(hashService.toHexString(hashService.createHash("someHash".getBytes())));
        attachmentTwoHash.setSignatureType(SignatureType.SHA_512);

        final var attachmentTwo = new ApiAttachment();
        attachmentTwo.setAttachmentId(UUID.randomUUID());
        attachmentTwo.setPurpose(Purpose.ATTACHMENT);
        attachmentTwo.setMimeType("application/pdf");
        attachmentTwo.setHash(attachmentTwoHash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(List.of(attachmentOne, attachmentTwo));

        final KeyPair keyPair = KeyGenerator.buildRSAKeyPair(4096);

        final JWK replyEncryptionKey = KeyGenerator.buildJWK(keyPair, "reply-key:12345678", KeyOperation.WRAP_KEY, JWEAlgorithm.RSA_OAEP_256);

        final ReplyChannel replyChannel = new ReplyChannel();
        replyChannel.setFitConnect(new FitConnect(EncryptionPublicKey.fromJwk(replyEncryptionKey), List.of("https://process.standard.org")));
        replyChannel.setDeMail(new DeMail("test@mail.de"));

        final var metadata = new Metadata();
        metadata.setSchema(config.getMetadataSchemaWriteVersion().toString());
        metadata.setContentStructure(contentStructure);
        metadata.setReplyChannel(replyChannel);

        final ServiceType serviceType = new ServiceType();
        serviceType.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");
        submission.setServiceType(serviceType);
        submission.setAttachments(contentStructure.getAttachments().stream().map(ApiAttachment::getAttachmentId).collect(Collectors.toList()));

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(submissionSchema));
        destinationService.setReplyChannels(replyChannel);

        final var destination = new Destination();
        destination.setServices(Set.of(destinationService));

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");
        authenticationTags.setAttachments(Map.of(attachmentOne.getAttachmentId(), "authTag", attachmentTwo.getAttachmentId(), "authTag"));

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertFalse(validationResult.isValid());
        assertThat(validationResult.getError().getMessage(), containsString("$.replyChannel: must have at most 1 properties"));
    }

    @Test
    void testIncorrectMetadataAuthenticationTag() {

        // Given
        final var metadata = new Metadata();
        final var destination = new Destination();

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("wrongMetadataAuthTag");

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), is(new IncorrectMetadataAuthenticationTag()));
    }


    @Test
    void testUnsupportedMetadataSchema() {

        // Given
        final String unsupportedSchema = "https://schema.fitko.de/fit-connect/metadata/1.0.0/metadata.schema.json/metadata_schema_3.2.1.json";

        final Data data = new Data();

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);
        metadata.setSchema(unsupportedSchema);
        final var destination = new Destination();

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), is(new UnsupportedMetadataSchema(unsupportedSchema)));
    }

    @Test
    void testMetadataSyntaxViolation() {

        // Given
        final var destination = new Destination();

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(null, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), instanceOf(MetadataJsonSyntaxViolation.class));
    }

    @Test
    void testMissingData() {

        // Given
        final var destination = new Destination();

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");

        final var metadata = new Metadata();
        metadata.setContentStructure(new ContentStructure());

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), instanceOf(MissingData.class));
    }

    @Test
    void testMetadataSchemaViolation() {

        // Given
        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setData(new Data());

        final var metadata = new Metadata();
        metadata.setSchema(config.getMetadataSchemaWriteVersion().toString());
        metadata.setContentStructure(contentStructure);

        final var destination = new Destination();

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), is(new MetadataSchemaViolation()));
    }

    @Test
    void testUnsupportedService() {

        // Given
        final SubmissionSchema submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fit-connect/metadata/1.0.0/metadata.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final Hash hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final Data data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(Collections.emptyList());
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setSchema(config.getMetadataSchemaWriteVersion().toString());
        metadata.setContentStructure(contentStructure);

        final DestinationService destinationService = new DestinationService();
        destinationService.setSubmissionSchemas(Set.of(submissionSchema));
        destinationService.setIdentifier("urn:de:fim:leika:leistung:9999999999999");

        final var destination = new Destination();
        destination.setServices(Set.of(destinationService));

        final ServiceType submissionServiceType = new ServiceType();
        submissionServiceType.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");
        submission.setServiceType(submissionServiceType);

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), instanceOf(UnsupportedService.class));
    }

    @Test
    void testUnsupportedDataSchema() {

        // Given
        final SubmissionSchema submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fit-connect/metadata/1.0.0/metadata.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final Hash hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final Data data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(Collections.emptyList());
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setSchema(config.getMetadataSchemaWriteVersion().toString());
        metadata.setContentStructure(contentStructure);

        final SubmissionSchema destinationSubmissionSchema = new SubmissionSchema();
        destinationSubmissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fit-connect/metadata/3.0.0/metadata.schema.json"));
        destinationSubmissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(destinationSubmissionSchema));

        final var destination = new Destination();
        destination.setServices(Set.of(destinationService));

        final ServiceType submissionServiceType = new ServiceType();
        submissionServiceType.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");
        submission.setServiceType(submissionServiceType);

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), instanceOf(UnsupportedDataSchema.class));
    }

    @Test
    void testAttachmentsMismatch() {

        // Given
        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://schema.fitko.de/fit-connect/metadata/1.0.0/metadata.schema.json"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Hash hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setSchema(config.getMetadataSchemaWriteVersion().toString());
        metadata.setContentStructure(contentStructure);

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setServices(Set.of(destinationService));

        final ServiceType submissionServiceType = new ServiceType();
        submissionServiceType.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");
        submission.setServiceType(submissionServiceType);

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");

        // attachmentId that were announced
        submission.setAttachments(List.of(UUID.randomUUID(), UUID.randomUUID(), UUID.randomUUID()));

        final Hash attachmentHash = new Hash();
        attachmentHash.setContent(hashService.toHexString(hashService.createHash("someHash".getBytes())));
        attachmentHash.setSignatureType(SignatureType.SHA_512);

        // attachment in metadata that doesn't match
        final ApiAttachment attachment = new ApiAttachment();
        attachment.setAttachmentId(UUID.randomUUID());
        attachment.setMimeType("application/pdf");
        attachment.setPurpose(Purpose.ATTACHMENT);
        attachment.setHash(attachmentHash);

        contentStructure.setAttachments(List.of(attachment));

        // authTag from submit event
        authenticationTags.setAttachments(Map.of(attachment.getAttachmentId(), "attachmentAuthTag"));

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), instanceOf(AttachmentsMismatch.class));
    }

    @Test
    void testUnsupportedReplyChannel() {

        // Given
        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://schema.fitko.de/fit-connect/metadata/1.0.0/metadata.schema.json"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Hash hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(Collections.emptyList());
        contentStructure.setData(data);

        final var metadata = new Metadata();
        config = new ApplicationConfig();
        metadata.setSchema(config.getMetadataSchemaWriteVersion().toString());
        metadata.setContentStructure(contentStructure);
        metadata.setReplyChannel(ReplyChannel.ofEmail("test@mail.org"));

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(schema));
        destinationService.setReplyChannels(ReplyChannel.ofDeMail("test@mail.de"));

        final var destination = new Destination();
        destination.setServices(Set.of(destinationService));

        final ServiceType submissionServiceType = new ServiceType();
        submissionServiceType.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");
        submission.setServiceType(submissionServiceType);

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), instanceOf(UnsupportedReplyChannel.class));
    }

    @Test
    void testMultipleDestinationServices() {

        // Given
        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://schema.fitko.de/fit-connect/metadata/1.2.0/metadata.schema.json"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Hash hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(Collections.emptyList());
        contentStructure.setData(data);

        final var metadata = new Metadata();
        config = new ApplicationConfig();
        metadata.setSchema(config.getMetadataSchemaWriteVersion().toString());
        metadata.setContentStructure(contentStructure);
        metadata.setReplyChannel(ReplyChannel.ofEmail("test@mail.org"));

        final DestinationService destinationService1 = new DestinationService();
        destinationService1.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService1.setSubmissionSchemas(Set.of(schema));
        destinationService1.setReplyChannels(ReplyChannel.ofFink("postBoxId", "https://fink.host.org"));

        final DestinationService destinationService2 = new DestinationService();
        destinationService2.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService2.setSubmissionSchemas(Set.of(schema));
        destinationService2.setReplyChannels(ReplyChannel.ofEmail("test@mail.org"));

        final var destination = new Destination();
        destination.setServices(Set.of(destinationService1, destinationService2));

        final ServiceType submissionServiceType = new ServiceType();
        submissionServiceType.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");
        submission.setServiceType(submissionServiceType);

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testDestinationWithNoReplyChannelAndSubmissionWithReplyChannel() {

        // Given
        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://schema.fitko.de/fit-connect/metadata/1.2.0/metadata.schema.json"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Hash hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(Collections.emptyList());
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setSchema(config.getMetadataSchemaWriteVersion().toString());
        metadata.setContentStructure(contentStructure);

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setServices(Set.of(destinationService));

        final ServiceType submissionServiceType = new ServiceType();
        submissionServiceType.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final var submission = new Submission();
        submission.setEncryptedMetadata("header.encryption_key.init_vector.ciphertext.metadataAuthTag");
        submission.setServiceType(submissionServiceType);

        final var authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTag");

        final ReplyChannel submissionReplyChannel = new ReplyChannel();
        submissionReplyChannel.setEMail(new Email("test@mail.org", false, null));
        metadata.setReplyChannel(submissionReplyChannel);

        // When
        final ValidationResult validationResult = underTest.validateSubmissionMetadata(metadata, submission, destination, authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems(), contains(new UnsupportedReplyChannel()));
    }

    @Test
    void validateMetadataWithoutSchemaAttributeButValidSchema() {

        final var contentStructure = new ContentStructure();
        contentStructure.setAttachments(Collections.emptyList());

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        assertTrue(underTest.validateMetadataSchema(metadata).isValid());
    }

    @Test
    void testMetadataWithInvalidSchemaAttribute() {

        final Metadata metadata = new Metadata();
        metadata.setSchema("https://schema.fitko.de/fit-connect/metadata/2.0.0/metadata.schema.json");

        final ValidationResult validationResult = underTest.validateMetadataSchema(metadata);

        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems(), contains(new UnsupportedMetadataSchema(metadata.getSchema())));
    }

    @Test
    void testMetadataWithMissingProperties() {

        // Given
        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var dataWithoutHashedContent = new Data();
        dataWithoutHashedContent.setSubmissionSchema(submissionSchema);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(dataWithoutHashedContent);
        contentStructure.setAttachments(Collections.emptyList());

        final var metadata = new Metadata();
        metadata.setSchema("https://schema.fitko.de/fit-connect/metadata/1.2.3/metadata.schema.json");
        metadata.setContentStructure(contentStructure);

        // When
        final ValidationResult validationResult = underTest.validateMetadataSchema(metadata);

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("$.contentStructure.data: required property 'hash' not found"));
    }

    @Test
    void testMetadataWithInvalidPattern() {

        // Given
        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var hash = new Hash();
        hash.setContent("noHexString");
        hash.setSignatureType(SignatureType.SHA_512);

        final var dataWithInvalidHashFormat = new Data();
        dataWithInvalidHashFormat.setSubmissionSchema(submissionSchema);
        dataWithInvalidHashFormat.setHash(hash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(dataWithInvalidHashFormat);
        contentStructure.setAttachments(Collections.emptyList());

        final var metadata = new Metadata();
        metadata.setSchema("https://schema.fitko.de/fit-connect/metadata/1.2.3/metadata.schema.json");
        metadata.setContentStructure(contentStructure);

        // When
        final ValidationResult validationResult = underTest.validateMetadataSchema(metadata);

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("$.contentStructure.data.hash.content: does not match the regex pattern ^[a-f0-9]{128}$"));
    }

    @Test
    void testIncorrectAttachmentAuthenticationTag() {

        // Given
        final var firstAttachmentId = UUID.randomUUID();
        final var secondAttachmentId = UUID.randomUUID();

        final var firstDecryptedData = "first test attachment".getBytes();
        final var firstEncryptedData = "part1.part2.part3.part4.authTag1";
        final var firstOriginalHash = hashService.toHexString(hashService.createHash(firstDecryptedData));

        final byte[] secondDecryptedData = "second test attachment".getBytes();
        final var secondEncryptedData = "part1.part2.part3.part4.authTag2";
        final var secondOriginalHash = hashService.toHexString(hashService.createHash(secondDecryptedData));

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setAttachments(Map.of(firstAttachmentId, "notMatchingAuthTag1", secondAttachmentId, "notMatchingAuthTag2"));

        final Hash firstHash = new Hash();
        firstHash.setContent(firstOriginalHash);

        final ApiAttachment firstAttachmentMetadata = new ApiAttachment();
        firstAttachmentMetadata.setAttachmentId(firstAttachmentId);
        firstAttachmentMetadata.setHash(firstHash);

        final Hash secondHash = new Hash();
        secondHash.setContent(secondOriginalHash);

        final ApiAttachment secondAttachmentMetadata = new ApiAttachment();
        secondAttachmentMetadata.setAttachmentId(secondAttachmentId);
        secondAttachmentMetadata.setHash(secondHash);

        final AttachmentForValidation firstAttachment = new AttachmentForValidation(firstAttachmentMetadata, firstEncryptedData, firstDecryptedData);
        final AttachmentForValidation secondAttachment = new AttachmentForValidation(secondAttachmentMetadata, secondEncryptedData, secondDecryptedData);

        // When
        final ValidationResult validationResult = underTest.validateAttachments(List.of(firstAttachment, secondAttachment), authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems(), hasSize(2));
        assertThat(validationResult.getProblems().stream().map(Problem::getClass).collect(Collectors.toList()), hasItems(IncorrectAttachmentAuthenticationTag.class));
        assertThat(validationResult.getProblems().get(0).getInstance(), containsString("attachment:" + firstAttachmentId));
        assertThat(validationResult.getProblems().get(1).getInstance(), containsString("attachment:" + secondAttachmentId));
    }

    @Test
    void testAttachmentHashMismatch() {

        // Given
        final var firstAttachmentId = UUID.randomUUID();
        final var secondAttachmentId = UUID.randomUUID();

        final var firstDecryptedData = "first test attachment".getBytes();
        final var firstEncryptedData = "part1.part2.part3.part4.authTag1";
        final var firstOriginalHash = hashService.toHexString(hashService.createHash(firstDecryptedData));

        final byte[] secondDecryptedData = "second test attachment".getBytes();
        final var secondEncryptedData = "part1.part2.part3.part4.authTag2";
        final var secondOriginalHash = hashService.toHexString(hashService.createHash(secondDecryptedData));

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setAttachments(Map.of(firstAttachmentId, "authTag1", secondAttachmentId, "authTag2"));

        final Hash firstHash = new Hash();
        firstHash.setContent(firstOriginalHash);

        final ApiAttachment firstAttachmentMetadata = new ApiAttachment();
        firstAttachmentMetadata.setAttachmentId(firstAttachmentId);
        firstAttachmentMetadata.setHash(firstHash);

        final Hash secondHash = new Hash();
        secondHash.setContent(secondOriginalHash);

        final ApiAttachment secondAttachmentMetadata = new ApiAttachment();
        secondAttachmentMetadata.setAttachmentId(secondAttachmentId);
        secondAttachmentMetadata.setHash(secondHash);

        final AttachmentForValidation firstAttachment = new AttachmentForValidation(firstAttachmentMetadata, firstEncryptedData, "modified attachment 1 content".getBytes());
        final AttachmentForValidation secondAttachment = new AttachmentForValidation(secondAttachmentMetadata, secondEncryptedData, "modified attachment 2 content".getBytes());

        // When
        final ValidationResult validationResult = underTest.validateAttachments(List.of(firstAttachment, secondAttachment), authenticationTags);

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems(), hasSize(2));
        assertThat(validationResult.getProblems().stream().map(Problem::getClass).collect(Collectors.toList()), hasItems(AttachmentHashMismatch.class));
        assertThat(validationResult.getProblems().get(0).getInstance(), containsString("attachment:" + firstAttachmentId));
        assertThat(validationResult.getProblems().get(1).getInstance(), containsString("attachment:" + secondAttachmentId));
    }

    @Test
    void testValidJsonData() {

        // Given
        final SchemaProvider schemaProvider = mock(SchemaProvider.class);
        when(schemaProvider.loadSubmissionDataSchema(any())).thenReturn(getResourceAsString("/submission-data-test-schema/submission_data_schema_remote.json"));

        final ApplicationConfig applicationConfig = getApplicationConfig(true, false);
        final DefaultValidationService defaultValidationService = new DefaultValidationService(applicationConfig, new HashService(), schemaProvider, trustedRootCertificates);

        final var decryptedData = "{ \"someString\": \"someString\", \"someArray\": [\"someString\"]}".getBytes();

        final var submission = new Submission();
        submission.setEncryptedData("part1.part2.part3.part4.dataAuthTag");

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash(decryptedData)));
        hash.setSignatureType(SignatureType.SHA_512);

        final var data = new Data();
        data.setHash(hash);
        data.setSubmissionSchema(submissionSchema);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(Collections.emptyList());

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        // When
        final ValidationResult validationResult = defaultValidationService.validateData(decryptedData, submission.getEncryptedData(), metadata, "dataAuthTag");

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testValidXmlData() {

        // Given
        final var decryptedXmlData = "<xml>test</xml>".getBytes();

        final var submission = new Submission();
        submission.setEncryptedData("part1.part2.part3.part4.dataAuthTag");

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("urn:de:fim:leika:leistung:99400048079000"));
        submissionSchema.setMimeType(MimeType.APPLICATION_XML);

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash(decryptedXmlData)));
        hash.setSignatureType(SignatureType.SHA_512);

        final var data = new Data();
        data.setHash(hash);
        data.setSubmissionSchema(submissionSchema);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(Collections.emptyList());

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        // When
        final ValidationResult validationResult = underTest.validateData(decryptedXmlData, submission.getEncryptedData(), metadata, "dataAuthTag");

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testIncorrectDataAuthenticationTag() {

        // Given
        final var decryptedData = "{ \"fachdaten\" : \"test\"}".getBytes();

        final var submission = new Submission();
        submission.setEncryptedData("part1.part2.part3.part4.wrongDataAuthTag");

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash(decryptedData)));
        hash.setSignatureType(SignatureType.SHA_512);

        final var data = new Data();
        data.setHash(hash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        // When
        final ValidationResult validationResult = underTest.validateData(decryptedData, submission.getEncryptedData(), metadata, "dataAuthTag");

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), instanceOf(IncorrectDataAuthenticationTag.class));
    }

    @Test
    void testDataHashMismatch() {

        // Given
        final var decryptedData = "{ \"fachdaten\" : \"test\"}".getBytes();

        final var submission = new Submission();
        submission.setEncryptedData("part1.part2.part3.part4.dataAuthTag");

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("manipulated data".getBytes())));
        hash.setSignatureType(SignatureType.SHA_512);

        final var data = new Data();
        data.setHash(hash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        // When
        final ValidationResult validationResult = underTest.validateData(decryptedData, submission.getEncryptedData(), metadata, "dataAuthTag");

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), instanceOf(DataHashMismatch.class));
    }

    @Test
    void testDataJsonSyntaxViolation() throws IOException {

        // Given
        final SchemaProvider schemaProvider = mock(SchemaProvider.class);
        when(schemaProvider.loadSubmissionDataSchema(any())).thenReturn(getResourceAsString("/submission-data-test-schema/submission_data_schema_remote.json"));

        final ApplicationConfig config = getApplicationConfig(true, false);
        final DefaultValidationService defaultValidationService = new DefaultValidationService(config, new HashService(), schemaProvider, trustedRootCertificates);

        final var decryptedData = "{ invalid_json }".getBytes();

        final var submission = new Submission();
        submission.setEncryptedData("part1.part2.part3.part4.dataAuthTag");

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash(decryptedData)));
        hash.setSignatureType(SignatureType.SHA_512);

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var data = new Data();
        data.setHash(hash);
        data.setSubmissionSchema(submissionSchema);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        // When
        final ValidationResult validationResult = defaultValidationService.validateData(
                decryptedData, submission.getEncryptedData(), metadata, "dataAuthTag");

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), instanceOf(DataJsonSyntaxViolation.class));
    }

    @Test
    void testDataXmlSyntaxViolation() {

        // Given
        final SchemaProvider schemaProvider = mock(SchemaProvider.class);
        when(schemaProvider.loadSubmissionDataSchema(any())).thenReturn(getResourceAsString("/submission-data-test-schema/submission_data_schema_remote.json"));

        final ApplicationConfig config = getApplicationConfig(true, false);
        final DefaultValidationService defaultValidationService = new DefaultValidationService(config, new HashService(), schemaProvider, trustedRootCertificates);

        final var decryptedData = "<xmehl>invalid xml</xml> }".getBytes();

        final var submission = new Submission();
        submission.setEncryptedData("part1.part2.part3.part4.dataAuthTag");

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash(decryptedData)));
        hash.setSignatureType(SignatureType.SHA_512);

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setMimeType(MimeType.APPLICATION_XML);

        final var data = new Data();
        data.setHash(hash);
        data.setSubmissionSchema(submissionSchema);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        // When
        final ValidationResult validationResult = defaultValidationService.validateData(
                decryptedData, submission.getEncryptedData(), metadata, "dataAuthTag");

        // Then
        assertTrue(validationResult.hasProblems());
        assertThat(validationResult.getProblems().get(0), instanceOf(DataXmlSyntaxViolation.class));
    }

    @Test
    void validateMatchingHash() {

        // Given
        final var hexHashFromSender = hashService.toHexString(hashService.createHash("some data".getBytes()));

        // When
        final ValidationResult validationResult = underTest.validateHashIntegrity(hexHashFromSender, "some data".getBytes());

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void validateCorruptHash() {

        // Given
        final var hexHashFromSender = "e1645e7492f032fb62c674db75500be7b260bfc0daa965821ddb3f8a49b5d33788ee3f04";

        // When
        final ValidationResult validationResult = underTest.validateHashIntegrity(hexHashFromSender, "corrupted data".getBytes());

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Metadata contains invalid hash value"));
    }

    @Test
    void testValidXml() {

        // Given
        final var validXml = getResourceAsString("/valid_xml_data.xml");

        // When
        final ValidationResult validationResult = underTest.validateXmlFormat(validXml);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testInvalidXml() {

        // Given
        final var invalidXml = getResourceAsString("/invalid_xml_data.xml");

        // When
        final ValidationResult validationResult = underTest.validateXmlFormat(invalidXml);

        // Then
        assertFalse(validationResult.isValid());
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("\"xsb:sender\" must be terminated by the matching end-tag \"</xsb:sender>\""));
    }

    @Test
    void testInvalidSetEvent() {

        // Given
        final var event = getResourceAsString("/invalid_test_set_event.json");

        // When
        final ValidationResult validationResult = underTest.validateSetEventSchema(event);

        // Then
        assertFalse(validationResult.isValid());
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("$: required property 'jti' not found"));
    }

    @Test
    void testValidSetEvent() {

        // Given
        final var event = getResourceAsString("/valid_test_set_event.json");

        // When
        final ValidationResult validationResult = underTest.validateSetEventSchema(event);

        // Then
        assertTrue(validationResult.isValid());
        assertFalse(validationResult.hasError());
    }

    @Test
    void testValidDestinationPayload() {

        // Given
        final Map<String, Object> claims = Map.of(
                "iss", "submission-service",
                "jti", UUID.randomUUID().toString(),
                "iat", new Date().getTime(),
                "destinationId", UUID.randomUUID().toString(),
                "submissionHost", "host",
                "services", Collections.emptyList()
        );

        // When
        final ValidationResult validationResult = underTest.validateDestinationSchema(claims);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testDestinationPayloadIsMissingMandatoryClaims() {

        // Given
        final Map<String, Object> claims = Map.of("test", "claim");

        // When
        final ValidationResult validationResult = underTest.validateDestinationSchema(claims);

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("required property 'jti' not found"));
        assertThat(validationResult.getError().getMessage(), containsString("required property 'iss' not found"));
        assertThat(validationResult.getError().getMessage(), containsString("required property 'iat' not found"));
        assertThat(validationResult.getError().getMessage(), containsString("required property 'services' not found"));
        assertThat(validationResult.getError().getMessage(), containsString("required property 'submissionHost' not found"));
        assertThat(validationResult.getError().getMessage(), containsString("required property 'destinationId' not found"));
    }

    @Test
    void testValidJson() {

        // When
        final ValidationResult validationResult = underTest.validateJsonFormat("{\"test\":\"test\"}");

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testInvalidJson() {

        // When
        final ValidationResult validationResult = underTest.validateJsonFormat("{test:\"test\"}");

        // Then
        assertFalse(validationResult.isValid());
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("was expecting double-quote to start field name"));
    }

    @Test
    void testValidSubmissionDataSchema() {

        // Given
        final SchemaProvider schemaProvider = mock(SchemaProvider.class);
        when(schemaProvider.loadSubmissionDataSchema(any())).thenReturn(getResourceAsString("/submission-data-test-schema/submission_data_schema_remote.json"));

        final ApplicationConfig config = getApplicationConfig(true, false);
        final DefaultValidationService defaultValidationService = new DefaultValidationService(config, new HashService(), schemaProvider, trustedRootCertificates);

        final var invalidJson = "{ \"someString\": \"someStringValue\", \"someArray\": [\"someArrayValue\"] }";

        // When
        final ValidationResult validationResult = defaultValidationService.validateSubmissionDataSchema(invalidJson, URI.create("urn:something"));

        // Then
        assertTrue(validationResult.isValid());
        assertFalse(validationResult.hasError());
    }

    @Test
    void testInvalidSubmissionDataSchema() {

        // Given
        final SchemaProvider schemaProvider = mock(SchemaProvider.class);
        when(schemaProvider.loadSubmissionDataSchema(any())).thenReturn(getResourceAsString("/submission-data-test-schema/submission_data_schema_remote.json"));

        final ApplicationConfig config = getApplicationConfig(true, false);
        final DefaultValidationService defaultValidationService = new DefaultValidationService(config, new HashService(), schemaProvider, trustedRootCertificates);

        final var invalidJson = "{ \"wrongString\": \"someStringValue\", \"someArray\": [\"someArrayValue\"] }";

        // When
        final ValidationResult validationResult = defaultValidationService.validateSubmissionDataSchema(invalidJson, URI.create("urn:something"));

        // Then
        assertFalse(validationResult.isValid());
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("$: required property 'someString' not found"));
    }

    @Test
    void testIgnoreInvalidSubmissionDataSchemaWhenValidationIsDeactivated() {

        // Given
        final SchemaProvider schemaProvider = mock(SchemaProvider.class);
        when(schemaProvider.loadSubmissionDataSchema(any())).thenReturn(getResourceAsString("/submission-data-test-schema/submission_data_schema_remote.json"));

        final ApplicationConfig config = getApplicationConfig(true, true);
        final DefaultValidationService defaultValidationService = new DefaultValidationService(config, new HashService(), schemaProvider, trustedRootCertificates);

        final var invalidJson = "{ \"wrongString\": \"someStringValue\", \"someArray\": [\"someArrayValue\"] }";

        // When
        final ValidationResult validationResult = defaultValidationService.validateSubmissionDataSchema(invalidJson, URI.create("urn:something"));

        // Then
        assertTrue(validationResult.isValid());
        assertFalse(validationResult.hasError());
    }

    @Test
    void testValidCallback() {

        final MessageDigestService mockedMessageDigestService = mock(MessageDigestService.class);
        when(mockedMessageDigestService.calculateHMAC(anyString(), anyString())).thenReturn("valid");

        final Environments testEnv = Environments.TEST;
        final ApplicationConfig config = ApplicationConfig.builder()
                .environments(Map.of(testEnv.getEnvironmentName(), testEnv.getEnvironment()))
                .activeEnvironment(testEnv.getEnvironmentName())
                .build();

        final DefaultValidationService defaultValidationService = new DefaultValidationService(
                config, mockedMessageDigestService, mock(SchemaProvider.class), Collections.emptyList());

        final ValidationResult validationResult = defaultValidationService.validateCallback(
                "valid", ZonedDateTime.now().toInstant().getEpochSecond(), "body", "secret");

        verify(mockedMessageDigestService, times(1)).calculateHMAC(anyString(), anyString());
        assertTrue(validationResult.isValid());
        assertFalse(validationResult.hasError());
    }

    @Test
    void testCallbackWithTimestampNotInSecondsDoesNotMatch() {

        // Given
        final var now = Instant.now();
        final var hmac = new HashService().calculateHMAC(now.getEpochSecond() + ".body", "callbackSecret");

        // When
        final ValidationResult validationResult = underTest.validateCallback(hmac, now.toEpochMilli(), "body", "callbackSecret");

        // Then
        assertFalse(validationResult.isValid());
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), is("HMAC provided by callback does not match the expected result."));

    }

    @Test
    void testCallbackWithExpiredTimestamp() {

        final ValidationResult validationResult = underTest.validateCallback("", 0L, "", "");

        assertFalse(validationResult.isValid());
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), is("Timestamp provided by callback is expired."));
    }

    @Test
    void testCallbackWithInvalidHmac() {

        final ValidationResult validationResult = underTest.validateCallback(
                "invalid", ZonedDateTime.now().toInstant().getEpochSecond(), "body", "secret");

        assertFalse(validationResult.isValid());
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), is("HMAC provided by callback does not match the expected result."));
    }

    @Test
    public void productiveTransmissionServiceCertificateIsValidAccordingToRootCertificates() throws ParseException, IOException {

        // Given
        final ApplicationConfig config = getApplicationConfig(false, false);
        final DefaultValidationService defaultValidationService = new DefaultValidationService(config, hashService, schemaProvider, trustedRootCertificates);

        final RSAKey rsaKey = RSAKey.parse(getResourceAsString("/certificates/grp-fit-connect-zustelldienst-produktivumgebung.json"));

        // When
        final ValidationResult validationResult = defaultValidationService.validatePublicKey(rsaKey, KeyOperation.VERIFY);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    public void fitConnectTestCertificateIsValidAccordingToTestRootCertificates() throws ParseException {

        // Given
        var config = getApplicationConfig(false, false);
        var trustedRootCert = getResourceAsString("/trusted-test-root-certificates/TEST-PCA20.pem");
        var defaultValidationService = new DefaultValidationService(config, hashService, schemaProvider, List.of(trustedRootCert));

        var rsaKey = RSAKey.parse(getResourceAsString("/certificates/grp-fitko-testzertifikat-fit-connect-1.json"));

        // When
        var validationResult = defaultValidationService.validatePublicKey(rsaKey, KeyOperation.WRAP_KEY);

        // Then
        assertTrue(validationResult.isValid());

    }

    @Test
    public void revokedFitConnectTestCertificateIsInvalidAccordingToTestRootCertificates() throws ParseException {

        // Given
        var config = getApplicationConfig(false, false);
        var trustedRootCert = getResourceAsString("/trusted-test-root-certificates/TEST-PCA20.pem");
        var defaultValidationService = new DefaultValidationService(config, hashService, schemaProvider, List.of(trustedRootCert));

        var rsaKey = RSAKey.parse(getResourceAsString("/certificates/grp-fitko-testzertifikat-fit-connect-2.json"));

        // When
        var validationResult = defaultValidationService.validatePublicKey(rsaKey, KeyOperation.WRAP_KEY);

        //Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), is("JWK with id rRaGt35McGycD4MeRyHZSLFo7dda2PEJscBuPT1KJdE-wrapKey returned certificate status REVOKED for OCSP check"));
    }

    @Test
    public void fitConnectTestCertificateIsInvalidAccordingToRootCertificates() throws ParseException {

        // Given
        final ApplicationConfig config = getApplicationConfig(false, false);
        final DefaultValidationService defaultValidationService = new DefaultValidationService(config, hashService, schemaProvider, trustedRootCertificates);

        final RSAKey rsaKey = RSAKey.parse(getResourceAsString("/certificates/grp-fitko-testzertifikat-fit-connect-1.json"));

        //When
        final ValidationResult validationResult = defaultValidationService.validatePublicKey(rsaKey, KeyOperation.WRAP_KEY);

        //Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), is("JWK with id LM0FPR9i-Yg1Cks-f_HG4dHocwLc2MU3rigME-Uc1Lc-wrapKey has invalid certificate chain"));
    }

    private String getResourceAsString(final String filename) throws RuntimeException {
        return new String(getResourceAsByteArray(filename));
    }

    private byte[] getResourceAsByteArray(final String filename) throws RuntimeException {
        try {
            return DefaultValidationServiceTest.class.getResourceAsStream(filename).readAllBytes();
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    private static ApplicationConfig getApplicationConfig(final boolean allowInsecureKey, final boolean skipSubmissionDataValidation) {
        final var envName = new EnvironmentName("testing");
        final var testing = new Environment();
        testing.setAllowInsecurePublicKey(allowInsecureKey);
        testing.setSkipSubmissionDataValidation(skipSubmissionDataValidation);
        return ApplicationConfig.builder()
                .environments(Map.of(envName, testing))
                .activeEnvironment(envName)
                .build();
    }

}